Imports System.Threading

Namespace System.Web
	'//////////////////////////////////////////////////
	' A delegate's BeginInvoke runs on a "worker thread" inside the CLR ThreadPool.
	' This allows you to run a delegate on a "completion port thread" inside the CLR ThreadPool
	' instead of a "worker thread" by calling IOBeginInvoke instead of BeginInvoke.
	' In a nut shell, this mechanism basically skips the actual system I/O and simply posts a
	' completion delegate to the completion port queue so the completion delegate will be 
	' executed by a "completion port thread" that is working the completion port queue.
	'
	' Example:
	' delegate.BeginInvoke => executes delegate on "worker thread".
	' delegate.IOBeginInvoke => executes delegate on "completion port thread".
	' 
	'
	' Extremely simplified explanation:
	'
	' CLR ThreadPool is made up of two pools of threads: "worker threads" and "completion port threads".
	'
	' Basically you can either queue a user work item which runs the delegate on a "worker thread",
	' or queue a native overlapped item which runs the completion delegate on a "completion port thread".
	' ThreadPool.QueueUserWorkItem (and delegate.BeginInvoke) => executes delegate on "worker thread".
	' ThreadPool.UnsafeQueueNativeOverlapped => executes completion delegate on "completion port thread".
	'
	'             (CLR ThreadPool)
	'               /        \
	' worker threads      completion port threads
	'
	'  o o  oo  o  oo                    _____________post to queue using ThreadPool.UnsafeQueueNativeOverlapped
	'    o o oo o  o                    |             (i.e. PostQueuedCompletionStatus)
	'   o  o o  o o                     v
	'  oo o  oo  o o                  |  |
	'    o  oo oo o                   |  | <----------completion port queue (one per process)
	'                                 |__|
	'       ^                         oo o
	'       |                        o o oo <---------completion port threads (work the completion port queue)
	'       |                 (cpu)(cpu)(cpu)(cpu)                            (i.e. GetQueuedCompletionStatus in loop)
	'       |                               
	'       |                               ^
	'       |                               |
	'       |                               |
	'       |                               |
	'       |                               |
	'       |    Each individual completion delegate is given to the completion port queue to execute,
	'       |    and the "completion port threads" working the completion port queue execute the delegate.
	'       |    (This has much less risk of thread explosion because each delegate just gets posted to the
	'       |     completion port queue, instead of being given to its own individual thread. Basically
	'       |     the queue grows, not the threads.)
	'       |
	'       |    The completion delegate is supplied in the overlapped structure that is posted to the
	'       |    completion port queue.
	'       |
	'       |    Posting to queue (PostQueuedCompletionStatus) => done by ThreadPool.UnsafeQueueNativeOverlapped.
	'       |    Working queue (GetQueuedCompletionStatus in loop) => done by "completion port thread" working queue.
	'       |
	'       |
	' Each individual delegate is given to its own individual "worker thread" to execute,
	' and the OS schedules the "worker thread" to run on a cpu.
	' (This has the risk of thread explosion because each new delegate executes on its own
	'  individual "worker thread". If the number of threads grows to large the entire OS can
	'  grind to a hault, with all the memory used and the cpu's spending all their time
	'  just trying to schedule the threads to run but not actually doing any work.)
	'
	'//////////////////////////////////////////////////
	Public Module IOThread
		Public Delegate Sub ProcessRequestDelegate(ByVal context As HttpContext)

		Public Class IOAsyncResult
			Implements IAsyncResult

			Public Property AsyncState As Object Implements IAsyncResult.AsyncState

			Public Property AsyncWaitHandle As WaitHandle Implements IAsyncResult.AsyncWaitHandle

			Public Property AsyncDelegate As [Delegate]

			Public Property CompletedSynchronously As Boolean Implements IAsyncResult.CompletedSynchronously

			Public Property IsCompleted As Boolean Implements IAsyncResult.IsCompleted

			Public Property Exception As Exception
		End Class

'ORIGINAL LINE: unsafe public static IAsyncResult IOBeginInvoke(Me ProcessRequestDelegate value, HttpContext context, AsyncCallback callback, object data)
		<System.Runtime.CompilerServices.Extension>
		Public Function IOBeginInvoke(ByVal value As ProcessRequestDelegate, ByVal context As HttpContext, ByVal callback As AsyncCallback, ByVal data As Object) As IAsyncResult
			Dim evt As New ManualResetEvent(False)

			Dim ar As New IOAsyncResult
			ar.AsyncState = New Object() { context, callback, data }
			ar.AsyncDelegate = value
			ar.AsyncWaitHandle = evt

			Dim o As New Overlapped(0, 0, IntPtr.Zero, ar)
			NativeOverlapped* no = o.Pack(New IOCompletionCallback(AddressOf ProcessRequestCompletionCallback), data)
			ThreadPool.UnsafeQueueNativeOverlapped(no)

			Return ar
		End Function

'ORIGINAL LINE: unsafe private static void ProcessRequestCompletionCallback(uint errorCode, uint numBytes, NativeOverlapped* no)
		Private Sub ProcessRequestCompletionCallback(ByVal errorCode As UInteger, ByVal numBytes As UInteger, NativeOverlapped* ByVal no As )
			Try
				Dim o As Overlapped = Overlapped.Unpack(no)

				Dim d As ProcessRequestDelegate = CType((CType(o.AsyncResult, IOAsyncResult)).AsyncDelegate, ProcessRequestDelegate)

				Dim state() As Object = CType(o.AsyncResult.AsyncState, Object)
				Dim context As HttpContext = CType(state(0), HttpContext)
				Dim callback As AsyncCallback = CType(state(1), AsyncCallback)
				Dim data As Object = state(2)

				Try
					d(context)
				Catch ex As Exception
					CType(o.AsyncResult, IOAsyncResult).Exception = ex
				End Try

				CType(o.AsyncResult, IOAsyncResult).IsCompleted = True
				CType(o.AsyncResult.AsyncWaitHandle, ManualResetEvent).Set

				If callback IsNot Nothing Then
					callback(o.AsyncResult)
				End If
			Finally
				Overlapped.Free(no)
			End Try
		End Sub

'ORIGINAL LINE: unsafe public static void IOEndInvoke(Me ProcessRequestDelegate value, IAsyncResult result)
		<System.Runtime.CompilerServices.Extension>
		Public Sub IOEndInvoke(ByVal value As ProcessRequestDelegate, ByVal result As IAsyncResult)
			Dim ar As IOAsyncResult = CType(result, IOAsyncResult)
			ar.AsyncWaitHandle.WaitOne(Timeout.Infinite)
			If ar.Exception IsNot Nothing Then
				Throw ar.Exception
			End If
		End Sub
	End Module
End Namespace
